"""
Create a function to transform StellarCursor data to RDD
"""

from __future__ import absolute_import
import abc
from future.utils import with_metaclass
import logging

try:
    import pyspark
    from pyspark import RDD, SparkContext
    from pyspark.serializers import BatchedSerializer
except ImportError:
    pyspark = None

_logger = logging.getLogger(__name__)


def transformToRDD(cursor, sc, parallelism=1):
    """
    Transform a StellarCursor to a Python RDD object

    param cursor: StellarCursor
    param sc: SparkContext
    param parallelism: Parallelism of RDD
    """
    if not pyspark:
        raise ImportError("Could not import pyspark! Please run `pip install pyspark` first in your environment!")

    # Get all data from cursor
    data = cursor.fetchall()

    # Set parallelism
    parallelism = max(1, parallelism)

    def reader_func(temp_filename):
        return sc._jvm.PythonRDD.readRDDFromFile(sc._jsc, temp_filename, parallelism)

    def createRDDServer():
        return sc._jvm.PythonParallelizeServer(sc._jsc.sc(), parallelism)

    batchSize = max(1, min(len(data) // parallelism, 1024))
    serializer = BatchedSerializer(sc._unbatched_serializer, batchSize)

    jrdd = sc._serialize_to_jvm(data, serializer, reader_func, createRDDServer)

    return RDD(jrdd, sc, serializer)



"""
Define database class for operations on a single graph.
"""

from __future__ import absolute_import
from types import TracebackType
import logging
import json
from typing import Any, Union

from ..graph_types import GraphSchema
from .connection import Connection, CypherResult

_logger = logging.getLogger(__name__)

class Graph(object):
    """
    Database class for operations on a single graph.
    """

    def __init__(self, graph_name: str, connection: Connection):
        self.graph_name = graph_name
        self.connection = connection
        self.graph_schema: GraphSchema = None
        self._init_connection()

    def __enter__(self):
        """
        Enter the database context.
        """
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the database context.
        """
        if self.connection:
            self.connection.close()
        self.connection = None

    def _init_connection(self):
        """
        Initialize the connection to the database.
        """
        
        self.connection.execute('config query.lang cypher')
        
        """Try to bind cursor with graph name."""
        try:
            self.connection.execute(f'use graph {self.graph_name}')
            self._get_graph_schema()
        except Exception as e:
            _logger.debug(f"graph {self.graph_name} not found")

    def _get_graph_schema(self):
        """
        Get the graph schema.
        """
        query_result = self.connection.execute('DESCRIBE GRAPH {} RAW'.format(self.graph_name))
        schemaInJson = query_result.fetchone()[0]
        query_result.close()
        self.graph_schema = GraphSchema.parseSchemaFromJson(schemaInJson)
        
        # get schema from data
        try:
            query_result = self.connection.execute('manipulate graph {} get_schema_from_data'.format(self.graph_name))
            self.graph_labels = query_result.fetchone()[0]
            query_result.close()
            self.graph_labels = json.loads(self.graph_labels)
        except:
            pass

    def execute(self, query: str, *args, **kwargs) -> CypherResult:
        """
        Execute a query on the database.
        """
        query_result = self.connection.execute(query, self.graph_schema)

        if (query.strip().lower().startswith('create graph')):
            self._get_graph_schema()
            
        return query_result

    def create(self, cypher: str):
        """
        Create a graph.
        """
        create_result = self.connection.execute(cypher)
        create_result.close()
        self._get_graph_schema()
    
    def drop(self):
        """
        Drop the graph.
        """
        drop_result = self.connection.execute('DROP GRAPH {}'.format(self.graph_name))
        drop_result.close()
        self.graph_schema = None

    def get_graph_schema(self) -> Union[GraphSchema, None]:
        """
        Get the graph schema.
        """
        return self.graph_schema
    
    def get_labels(self) -> Any:
        """
        Get the labels of the graph.
        """
        return self.graph_labels


"""DB-API implementation backed by PyHive

See http://www.python.org/dev/peps/pep-0249/

Many docstrings in this file are based on the PEP, which is in the public domain.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from pyhive import hive
from . import graph_types
from . import stellar_rdd
from .sasl_compat import PureSASLClient

from TCLIService import TCLIService
from TCLIService import ttypes
# Make all exceptions visible in this module per DB-API
from pyhive.exc import *  # noqa
from builtins import range
import contextlib
import getpass
import logging
import sys
import json
import thrift.protocol.TBinaryProtocol
import thrift.transport.TSocket
import thrift.transport.TTransport

_logger = logging.getLogger(__name__)


def connect(*args, **kwargs):
    """Constructor for creating a connection to the database. See class :py:class:`Connection` for
    arguments.

    :returns: a :py:class:`Connection` object.
    """
    return StellarConnection(*args, **kwargs)


class StellarConnection(object):
    """Wraps a Thrift session"""
    def __init__(self,
                 host=None,
                 port=None,
                 username=None,
                 database='default',
                 auth=None,
                 configuration=None,
                 kerberos_service_name=None,
                 password=None,
                 thrift_transport=None,
                 graph_name=None):
        """Connect to HiveServer2

        :param host: What host HiveServer2 runs on
        :param port: What port HiveServer2 runs on. Defaults to 10000.
        :param auth: The value of hive.server2.authentication used by HiveServer2.
            Defaults to ``NONE``.
        :param configuration: A dictionary of Hive settings (functionally same as the `set` command)
        :param kerberos_service_name: Use with auth='KERBEROS' only
        :param password: Use with auth='LDAP' or auth='CUSTOM' only
        :param thrift_transport: A ``TTransportBase`` for custom advanced usage.
            Incompatible with host, port, auth, kerberos_service_name, and password.
        :param graph_name: The name of graph to manipulate.

        The way to support LDAP and GSSAPI is originated from cloudera/Impyla:
        https://github.com/cloudera/impyla/blob/255b07ed973d47a3395214ed92d35ec0615ebf62
        /impala/_thrift_api.py#L152-L160
        """
        username = username or getpass.getuser()
        configuration = configuration or {}

        if (password is not None) != (auth in ('LDAP', 'CUSTOM')):
            raise ValueError(
                "Password should be set if and only if in LDAP or CUSTOM mode; "
                "Remove password or use one of those modes")
        if (kerberos_service_name is not None) != (auth == 'KERBEROS'):
            raise ValueError(
                "kerberos_service_name should be set if and only if in KERBEROS mode"
            )
        if thrift_transport is not None:
            has_incompatible_arg = (host is not None or port is not None
                                    or auth is not None
                                    or kerberos_service_name is not None
                                    or password is not None)
            if has_incompatible_arg:
                raise ValueError(
                    "thrift_transport cannot be used with "
                    "host/port/auth/kerberos_service_name/password")

        if thrift_transport is not None:
            self._transport = thrift_transport
        else:
            if port is None:
                port = 10000
            if auth is None:
                auth = 'NONE'
            socket = thrift.transport.TSocket.TSocket(host, port)
            if auth == 'NOSASL':
                # NOSASL corresponds to hive.server2.authentication=NOSASL in hive-site.xml
                self._transport = thrift.transport.TTransport.TBufferedTransport(
                    socket)
            elif auth in ('LDAP', 'KERBEROS', 'NONE', 'CUSTOM'):
                import thrift_sasl

                if auth == 'KERBEROS':
                    # KERBEROS mode in hive.server2.authentication is GSSAPI in SASL
                    sasl_auth = 'GSSAPI'
                else:
                    sasl_auth = 'PLAIN'
                    if password is None:
                        # Password doesn't matter in NONE mode, just needs to be nonempty.
                        password = 'x'

                def sasl_factory():
                    if sasl_auth == 'GSSAPI':
                        sasl_client = PureSASLClient(host, mechanism="GSSAPI", service=kerberos_service_name)
                    elif sasl_auth == 'PLAIN':
                        sasl_client = PureSASLClient(host, mechanism="PLAIN", username=username, password=password)
                    else:
                        raise AssertionError("Unsupported SASL mechanism")
                    return sasl_client

                self._transport = thrift_sasl.TSaslClientTransport(sasl_factory, sasl_auth, socket)
            else:
                # All HS2 config options:
                # https://cwiki.apache.org/confluence/display/Hive/Setting+Up+HiveServer2#SettingUpHiveServer2-Configuration
                # PAM currently left to end user via thrift_transport option.
                raise NotImplementedError(
                    "Only NONE, NOSASL, LDAP, KERBEROS, CUSTOM "
                    "authentication are supported, got {}".format(auth))

        protocol = thrift.protocol.TBinaryProtocol.TBinaryProtocol(
            self._transport)
        self._client = TCLIService.Client(protocol)
        # oldest version that still contains features we care about
        # "V6 uses binary type for binary payload (was string) and uses columnar result set"
        protocol_version = ttypes.TProtocolVersion.HIVE_CLI_SERVICE_PROTOCOL_V6

        self._graph_schema = None
        self._graph_schema_from_data = None

        try:
            self._transport.open()
            open_session_req = ttypes.TOpenSessionReq(
                client_protocol=protocol_version,
                configuration=configuration,
                username=username,
            )
            response = self._client.OpenSession(open_session_req)
            hive._check_status(response)
            assert response.sessionHandle is not None, "Expected a session from OpenSession"
            self._sessionHandle = response.sessionHandle
            assert response.serverProtocolVersion == protocol_version, \
                "Unable to handle protocol version {}".format(response.serverProtocolVersion)

            # parse schema if the graph_name is assigned
            if graph_name is not None:
                with contextlib.closing(self.cursor()) as cursor:
                    cursor.execute('set query.lang=cypher')
                    cursor.execute('DESCRIBE GRAPH {} RAW'.format(graph_name))

                    schemaInJson = cursor.fetchone()[0]

                    self._graph_schema = graph_types.GraphSchema.parseSchemaFromJson(
                        schemaInJson)
                    
                    # get schema from data
                    try:
                        cursor.execute('manipulate graph {} get_schema_from_data'.format(graph_name))
                        self._graph_schema_from_data = cursor.fetchone()[0]
                        self._graph_schema_from_data = json.loads(self._graph_schema_from_data)
                    except:
                        pass
            else:
                assert response.serverProtocolVersion == protocol_version, \
                    "Unable to handle protocol version {}".format(response.serverProtocolVersion)
                with contextlib.closing(self.cursor()) as cursor:
                    cursor.execute('set query.lang=sql')
                    cursor.execute('USE `{}`'.format(database))
        except:
            self._transport.close()
            raise

    def __enter__(self):
        """Transport should already be opened by __init__"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Call close"""
        self.close()

    def close(self):
        """Close the underlying session and Thrift transport"""
        req = ttypes.TCloseSessionReq(sessionHandle=self._sessionHandle)
        response = self._client.CloseSession(req)
        self._transport.close()
        hive._check_status(response)

    def commit(self):
        """Hive does not support transactions, so this does nothing."""
        pass

    def cursor(self, *args, **kwargs):
        """Return a new :py:class:`Cursor` object using the connection."""
        return StellarCursor(self, *args, **kwargs)

    @property
    def client(self):
        return self._client

    @property
    def sessionHandle(self):
        return self._sessionHandle

    def rollback(self):
        raise NotSupportedError(
            "Hive does not have transactions")  # pragma: no cover

    def getGraphSchema(self):
        return self._graph_schema

    def getSchemaFromData(self) -> dict:
        """
        Get schema from graph data.
        The difference of getGraphSchema() and getSchemaFromData() is the latter one will return start node labels and end node labels of edges.
        Result format is:
        {
            "vertices": [
                { "label": "label_a", "fields": ["prop1", "prop2", ...] }
            ],
            "edges": [
                "label": "label_b",
                "fields": ["prop3", "prop4", ...],
                "src_dst_labels": [
                    {"src": "label_x", "dst": "label_y"},
                ]
            ]
        }
        """
        return self._graph_schema_from_data

class StellarCursor(hive.Cursor):
    """These objects represent a database cursor, which is used to manage the context of a fetch
    operation.

    Cursors are not isolated, i.e., any changes done to the database by a cursor are immediately
    visible by other cursors or connections.
    """
    def __init__(self, connection, arraysize=1000):
        super(StellarCursor, self).__init__(connection, arraysize)
        self._column_comments = []
        self._graph_schema = connection.getGraphSchema()

    @property
    def description(self):
        """This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences contains information describing one result column:

        - name
        - type_code
        - display_size (None in current implementation)
        - internal_size (None in current implementation)
        - precision (None in current implementation)
        - scale (None in current implementation)
        - null_ok (always True in current implementation)

        This attribute will be ``None`` for operations that do not return rows or if the cursor has
        not had an operation invoked via the :py:meth:`execute` method yet.

        The ``type_code`` can be interpreted by comparing it to the Type Objects specified in the
        section below.
        """
        if self._operationHandle is None or not self._operationHandle.hasResultSet:
            return None
        if self._description is None:
            req = ttypes.TGetResultSetMetadataReq(self._operationHandle)
            response = self._connection.client.GetResultSetMetadata(req)
            hive._check_status(response)
            columns = response.schema.columns
            self._description = []
            # If it's a cypher query, column comment is not null
            self._column_comments = [
                col.comment for col in response.schema.columns
                if col.comment is not None
            ]

            for col in columns:
                primary_type_entry = col.typeDesc.types[0]
                if primary_type_entry.primitiveEntry is None:
                    # All fancy stuff maps to string
                    type_code = ttypes.TTypeId._VALUES_TO_NAMES[
                        ttypes.TTypeId.STRING_TYPE]
                else:
                    type_id = primary_type_entry.primitiveEntry.type
                    type_code = ttypes.TTypeId._VALUES_TO_NAMES[type_id]
                self._description.append(
                    (col.columnName.decode('utf-8')
                     if sys.version_info[0] == 2 else col.columnName,
                     type_code.decode('utf-8') if sys.version_info[0] == 2 else
                     type_code, None, None, None, None, True))
        return self._description

    def fetchone(self):
        row = super(StellarCursor, self).fetchone()

        # Empty data / graph name not assigned / Not a cypher query
        if row is None or self._graph_schema is None or len(
                self._column_comments) == 0:
            return row
        else:
            parsed_row = []

            for i in range(0, len(self._column_comments)):
                parsed_row.append(
                    self._convertData(self._column_comments[i], row[i]))

            return tuple(parsed_row)

    def _convertData(self, type, data):
        """Convert Crux type to Readable type"""
        if type == 'boolean':
            return bool(data)
        elif type == 'int':
            return int(data)
        elif type == 'long':
            return int(data)
        elif type == 'float' or type == 'double':
            return float(data)
        elif type == 'CruxType:Node' or type == 'GraphNode':
            return graph_types.Vertex.parseVertexFromJson(data)
        elif type == 'CruxType:Relation' or type == 'GraphRelation':
            return graph_types.Edge.parseEdgeFromJson(self._graph_schema, data)
        elif type == 'CruxType:Path':
            return graph_types.Path.parsePathFromJson(self._graph_schema, data)
        elif type.startswith('CruxType:List'):
            return self._parseList(type, data)
        elif type.startswith('CruxType:Map'):
            return self._parseMap(type, data)
        else:
            return data

    def _parseList(self, type, data):
        """Parse 'CruxType:List' type"""
        parsed_data = json.loads(data)
        newType = type[len('CruxType:List') + 1:type.find('>')]

        return [self._convertData(newType, json.dumps(entry)) for entry in parsed_data]

    def _parseMap(self, type, data):
        """Parse 'CruxType:Map' type"""
        parsed_data = json.loads(data)
        newTypes = type[len('CruxType:Map') + 1:-2].split(',')

        result = {}

        for entry in parsed_data.keys():
            key = self._convertData(newTypes[0], entry)
            result[key] = self._convertData(newTypes[1], parsed_data[entry])

        return result

    def toRDD(self, sc, parallelism=1):
        """
        Transform to RDD
        param sc: SparkContext
        param parallelism: RDD parallelism
        """
        return stellar_rdd.transformToRDD(self, sc, parallelism)



from __future__ import absolute_import

from puresasl.client import SASLClient, SASLError
from contextlib import contextmanager

@contextmanager
def error_catcher(self, Exc = Exception):
    try:
        self.error = None
        yield
    except Exc as e:
        self.error = str(e)


class PureSASLClient(SASLClient):
    def __init__(self, *args, **kwargs):
        self.error = None
        super(PureSASLClient, self).__init__(*args, **kwargs)

    def start(self, mechanism):
        with error_catcher(self, SASLError):
            if isinstance(mechanism, list):
                self.choose_mechanism(mechanism)
            else:
                self.choose_mechanism([mechanism])
            return True, self.mechanism, self.process()
        # else
        return False, mechanism, None

    def encode(self, incoming):
        with error_catcher(self):
            return True, self.unwrap(incoming)
        # else
        return False, None

    def decode(self, outgoing):
        with error_catcher(self):
            return True, self.wrap(outgoing)
        # else
        return False, None

    def step(self, challenge=None):
        with error_catcher(self):
            return True, self.process(challenge)
        # else
        return False, None

    def getError(self):
        return self.error



"""
Define vertex, edge, path and schema data structure
"""

from __future__ import absolute_import
import abc
from future.utils import with_metaclass
import json
import logging
import binascii
from typing import cast

_logger = logging.getLogger(__name__)


class GraphElement(with_metaclass(abc.ABCMeta, object)):
    """Base class of vertex and edge"""
    def __init__(self, label):
        self._label = label
        self._fields = {}
        self._tags = []
        self._rowKeyHexString = None

    def getLabel(self):
        return self._label

    def getField(self, name):
        return self._fields[name]

    def hasField(self, name):
        return name in self._fields

    def getFields(self):
        return self._fields

    def setFeild(self, name, value):
        self._fields[name] = value

    def getTags(self):
        return self._tags

    def setTags(self, newTags):
        self._tags = newTags

    def setRowKeyHexString(self, rowkey):
        self._rowKeyHexString = rowkey

    def getRowKeyHexString(self):
        return self._rowKeyHexString
    
    def __hash__(self):
        if self._rowKeyHexString:
            return hash(self._rowKeyHexString)
        else:
            return hash(self._fields)

    def __eq__(self, value):
        if isinstance(value, GraphElement):
            if self._rowKeyHexString:
                return self._rowKeyHexString == value._rowKeyHexString
            else:
                return self._fields == value._fields
        else:
            return False


class Vertex(GraphElement):
    """
    Vertex consists of a user-defined id, a label, a number of properties, and a list of tags.
    Empty properties and tags will be execluded in JSON string.
    """
    def __init__(self, uid, label):
        super(Vertex, self).__init__(label)
        self._uid = uid

    def getUserId(self):
        return self._uid

    def toJSON(self):
        m = {
            'type': 'vertex',
            'label': self._label,
            'uid': self._uid,
            'RowKeyHexString': self._rowKeyHexString,
        }

        if self._tags is not None and len(self._tags) > 0:
            m['tags'] = self._tags

        if self._fields is not None and len(self._fields) > 0:
            m['properties'] = self._fields

        return m

    def __str__(self):
        return json.dumps(self.toJSON(), ensure_ascii=False)

    @staticmethod
    def parseVertexFromJson(json_str):
        """Parse a Vertex object from a JSON string"""
        m = json.loads(json_str)

        return Vertex.parseVertexFromDict(m)

    @staticmethod
    def parseVertexFromDict(m):
        """Parse a Vertex object from a dict"""
        if 'labels' not in m:
            raise ValueError("Could not find label in JSON")

        prop_dict = m['properties']

        if '__uid' not in prop_dict:
            raise ValueError("Could not find uid in JSON")

        vertex = Vertex(prop_dict['__uid'], m['labels'][0])

        for key in prop_dict.keys():
            if key != '__uid' and key != '__tags':
                vertex.setFeild(key, prop_dict[key])

        if '__tags' in prop_dict:
            vertex.setTags(prop_dict['__tags'])

        rk = " ".join(map(lambda x: str(x), m['entityKey']))
        vertex.setRowKeyHexString(rk)

        return vertex

    @staticmethod
    def parseUidFromRK(rk):
        """Parse user-defined id from vertex row key in byte array"""
        return bytearray([x & 0xFF for x in rk[:-2]]).decode('utf-8')

    @staticmethod
    def parseLabelIdxFromRK(rk):
        """Parse label index from vertex row key in byte array"""
        label_in_little_endian = rk[-2:]
        #reverse to big endian
        label_in_little_endian.reverse()
        return int(binascii.hexlify(bytearray(label_in_little_endian)), 16)

    @staticmethod
    def parseShardIdFromRKV18(rk):
        """Parse shard id from vertex row key in byte array for graphSchema V18"""
        shard_id = (rk[0] & 0xFF) << 8
        shard_id |= rk[1] & 0xF0
        return int(shard_id >> 4)

    @staticmethod
    def parseLabelIdxFromRKV18(rk):
        """Parse label index from vertex row key in byte array for graphSchema V18"""
        label_index = (rk[1] & 0x0F) << 8
        label_index |= rk[2] & 0xFF
        return int(label_index)

    @staticmethod
    def parseInnerIdFromRKV18(rk, offset):
        """Parse long type inner id from vertex row key in byte array for graphSchema V18"""
        ID_LEN = 8
        inner_id = rk[offset + ID_LEN - 1] & 0x00FF
        inner_id |= (rk[offset + ID_LEN - 2] & 0x00FF) << 8
        inner_id |= (rk[offset + ID_LEN - 3] & 0x00FF) << 16
        inner_id |= (rk[offset + ID_LEN - 4] & 0x00FF) << 24
        inner_id |= (rk[offset + ID_LEN - 5] & 0x00FF) << 32
        return int(inner_id)


class Edge(GraphElement):
    """
    Edge consists of a start node, an end node, a label, and a list of tags.
    If multiple edges exist with the same start node and end node, an extra user-defined edge id should be assigned.
    Empty properties and tags will be execluded in JSON string.
    """
    def __init__(self, label):
        super(Edge, self).__init__(label)
        self._startNode = None
        self._endNode = None
        self._uid = ""

    def getStartNode(self):
        return self._startNode

    def setStartNode(self, startNode):
        self._startNode = startNode

    def getEndNode(self):
        return self._endNode

    def setEndNode(self, endNode):
        self._endNode = endNode

    def getUid(self):
        return self._uid

    def setUid(self, uid):
        self._uid = uid

    def toJSON(self):
        m = {
            'type': 'edge',
            'label': self._label,
            'euid': self._uid,
            'startNode': self._startNode.toJSON(),
            'endNode': self._endNode.toJSON(),
            'RowKeyHexString': self._rowKeyHexString,
        }

        if self._tags is not None and len(self._tags) > 0:
            m['tags'] = self._tags

        if self._fields is not None and len(self._fields) > 0:
            m['properties'] = self._fields

        return m

    def __str__(self):
        return json.dumps(self.toJSON(), ensure_ascii=False)

    @staticmethod
    def parseEdgeFromJson(schema, json_str):
        """Parse an Edge object from a JSON string"""
        m = json.loads(json_str)

        return Edge.parseEdgeFromDict(schema, m)

    @staticmethod
    def parseEdgeFromDict(schema, m):
        """Parse an Edge object from a dict"""
        if 'labels' not in m:
            raise ValueError("Could not find label in JSON")

        edge = Edge(m['labels'][0])

        rk = " ".join(map(lambda x: str(x), m['entityKey']))
        edge.setRowKeyHexString(rk)

        prop_dict = m['properties']

        # parse start node
        if 'startKey' not in m:
            raise ValueError("Could not find start node entity key in JSON")

        if schema.getVersion() >= 18:
            startUid = prop_dict['__srcuid']
            startLabelIdx = Vertex.parseLabelIdxFromRKV18(m['startKey'])
        else:
            startUid = Vertex.parseUidFromRK(m['startKey'])
            startLabelIdx = Vertex.parseLabelIdxFromRK(m['startKey'])
        startLabel = schema.getVertexLabel(startLabelIdx)

        if startLabel is None:
            raise ValueError(
                'Could not find start node label with label index `{}`'.format(
                    startLabelIdx))
        
        start_node = Vertex(startUid, startLabel)
        start_node.setRowKeyHexString(" ".join(map(lambda x: str(x), m['entityKey'][:8])))
        edge.setStartNode(start_node)

        # parse end node
        if 'endKey' not in m:
            raise ValueError("Could not find end node entity key in JSON")

        if schema.getVersion() >= 18:
            endUid = prop_dict['__dstuid']
            endLabelIdx = Vertex.parseLabelIdxFromRKV18(m['endKey'])
        else:
            endUid = Vertex.parseUidFromRK(m['endKey'])
            endLabelIdx = Vertex.parseLabelIdxFromRK(m['endKey'])
        endLabel = schema.getVertexLabel(endLabelIdx)

        if endLabel is None:
            raise ValueError(
                'Could not find end node label with label index `{}`'.format(
                    endLabelIdx))

        end_node = Vertex(endUid, endLabel)
        end_node.setRowKeyHexString(" ".join(map(lambda x: str(x), m['entityKey'][8:16])))
        edge.setEndNode(end_node)

        # parse extra edge id
        if '__uid' in prop_dict:
            edge.setUid(prop_dict['__uid'])

        # parse properties
        for key in prop_dict.keys():
            if key != '__uid' and key != '__tags':
                edge.setFeild(key, m['properties'][key])

        # parse tags
        if '__tags' in prop_dict:
            edge.setTags(prop_dict['__tags'])

        return edge


class Path(object):
    """A Path object consists of a list of Vertex and Edge objects"""
    def __init__(self, elems):
        self._elems = elems

    def getElements(self):
        return self._elems

    def length(self):
        return len(self._elems)

    def __str__(self):
        return str([str(entry) for entry in self._elems])
    
    def __hash__(self):
        rks = []
        for elem in self._elems:
            rks.append(cast(GraphElement, elem).getRowKeyHexString())
        
        return hash(tuple(rks))

    def __eq__(self, value):
        if isinstance(value, Path):
            if self.length() == value.length():
                return self.getElements() == value.getElements()
            else:
                return False
        else:
            return False

    @staticmethod
    def parsePathFromJson(schema, json_str):
        m = json.loads(json_str)
        elems = []

        for i in range(0, len(m)):
            if i % 2 == 0:
                elems.append(Vertex.parseVertexFromDict(m[i]))
            else:
                elems.append(Edge.parseEdgeFromDict(schema, m[i]))

        return Path(elems)


class GraphSchema(object):
    """
    A schema of graph contains a list of meta info:
    schema_version: Version of schema
    graph_name: Name of graph, unique in a cluster
    shard_number: Shard number of graph
    index_type: Type of index, could only be 'native' now
    replication_number: Replication number of graph, could be 1, 3, or other odd integer
    """
    def __init__(self, schema_version, graph_name, shard_number, index_type,
                 replication_number):
        self._schema_version = schema_version
        self._graph_name = graph_name
        self._shard_number = shard_number
        self._index_type = index_type
        self._replication_number = replication_number
        self._vertex_schemas = []
        self._edge_schemas = []

        # since V8
        self._encryption_type = None
        self._index_separated = None

    def setVertexSchemas(self, schema):
        self._vertex_schemas = schema

    def setEdgeSchemas(self, schema):
        self._edge_schemas = schema

    def setEncryptionType(self, enc_type):
        self._encryption_type = enc_type

    def setIndexSeparate(self, idx_sep):
        self._index_separated = idx_sep

    def getVertexLabel(self, labelIndex):
        for schema in self._vertex_schemas:
            if schema['label.index'] == labelIndex:
                return schema['label.value']

        return None

    def getEdgeLabel(self, labelIndex):
        for schema in self._edge_schemas:
            if schema['label.index'] == labelIndex:
                return schema['label.value']

        return None

    def getVersion(self):
        return self._schema_version

    def toJSON(self):
        m = {
            '__VERSION': self._schema_version,
            'graph.name': self._graph_name,
            'graph.replication.number': self._replication_number,
            'graph.shard.number': self._shard_number,
            'graph.index.type': self._index_type,
            'vertex.tables': self._vertex_schemas,
            'edge.tables': self._edge_schemas,
        }

        # since V8
        if self._encryption_type:
            m['graph.encryption.type'] = self._encryption_type

        # since V8
        if self._index_separated:
            m['graph.index.separated'] = self._index_separated

        return m

    def __str__(self):
        return json.dumps(self.toJSON(), ensure_ascii=False)

    @staticmethod
    def parseSchemaFromJson(json_str):
        """Parse a schema from JSON string"""
        m = json.loads(json_str)

        if '__VERSION' not in m:
            raise ValueError('Could not find `__VERSION` in graph schema')

        if 'graph.index.type' not in m:
            raise ValueError(
                'Could not find `graph.index.type` in graph schema')

        if 'graph.name' not in m:
            raise ValueError('Could not find `graph.name` in graph schema')

        if 'graph.replication.number' not in m:
            raise ValueError(
                'Could not find `graph.replication.number` in graph schema')

        if 'graph.shard.number' not in m:
            raise ValueError(
                'Could not find `graph.shard.number` in graph schema')

        if 'vertex.tables' not in m:
            raise ValueError('Could not find `vertex.tables` in graph schema')

        if 'edge.tables' not in m:
            raise ValueError('Could not find `edge.tables` in graph schema')

        result = GraphSchema(m['__VERSION'], m['graph.name'], int(m['graph.shard.number']),
                             m['graph.index.type'],
                             int(m['graph.replication.number']))

        result.setVertexSchemas(m['vertex.tables'])
        result.setEdgeSchemas(m['edge.tables'])

        # since V8
        if 'graph.encryption.type' in m:
            result.setEncryptionType(m['graph.encryption.type'])

        # since V8
        if 'graph.index.separated' in m:
            result.setIndexSeparate(m['graph.index.separated'])

        return result




from ._version import get_versions
from .v2 import Connection, Graph, CypherResult


__version__ = get_versions()['version']
del get_versions

__all__ = [
    '__version__',
    'stellar_hive',
    'stellar_rdd',
    'graph_types',
    'Connection',
    'Graph',
    'CypherResult',
]


from __future__ import absolute_import
from __future__ import unicode_literals

from types import TracebackType
from typing import Any, Union
from pyhive import hive
from pyhive.exc import *
from TCLIService import TCLIService
from TCLIService import ttypes

import thrift.protocol.TBinaryProtocol
import thrift.transport.TSocket
import thrift.transport.TTransport
import getpass
import logging
import json
import sys

from pystellardb.stellar_rdd import transformToRDD
from pystellardb.graph_types import Vertex, Edge, Path, GraphSchema
from pystellardb.sasl_compat import PureSASLClient

__logger = logging.getLogger(__name__)

class CypherResult(hive.Cursor):
    """
    QueryResult class for handling query results.
    """

    def __init__(self, connection: Any, graph_schema: GraphSchema, arraysize: int = 1000):
        super(CypherResult, self).__init__(connection, arraysize)
        self._graph_schema = graph_schema
        self._column_comments = []

    @property
    def description(self):
        """This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences contains information describing one result column:

        - name
        - type_code
        - display_size (None in current implementation)
        - internal_size (None in current implementation)
        - precision (None in current implementation)
        - scale (None in current implementation)
        - null_ok (always True in current implementation)

        This attribute will be ``None`` for operations that do not return rows or if the cursor has
        not had an operation invoked via the :py:meth:`execute` method yet.

        The ``type_code`` can be interpreted by comparing it to the Type Objects specified in the
        section below.
        """
        if self._operationHandle is None or not self._operationHandle.hasResultSet:
            return None
        if self._description is None:
            req = ttypes.TGetResultSetMetadataReq(self._operationHandle)
            response = self._connection.client.GetResultSetMetadata(req)
            hive._check_status(response)
            columns = response.schema.columns
            self._description = []
            # If it's a cypher query, column comment is not null
            self._column_comments = [
                col.comment for col in response.schema.columns
                if col.comment is not None
            ]

            for col in columns:
                primary_type_entry = col.typeDesc.types[0]
                if primary_type_entry.primitiveEntry is None:
                    # All fancy stuff maps to string
                    type_code = ttypes.TTypeId._VALUES_TO_NAMES[
                        ttypes.TTypeId.STRING_TYPE]
                else:
                    type_id = primary_type_entry.primitiveEntry.type
                    type_code = ttypes.TTypeId._VALUES_TO_NAMES[type_id]
                self._description.append(
                    (col.columnName.decode('utf-8')
                     if sys.version_info[0] == 2 else col.columnName,
                     type_code.decode('utf-8') if sys.version_info[0] == 2 else
                     type_code, None, None, None, None, True))
        return self._description
    
    def fetchone(self):
        row = super(CypherResult, self).fetchone()

        if row is None:
            return None

        parsed_row = []
        for i in range(0, len(self._column_comments)):
            parsed_row.append(
                self._convertData(self._column_comments[i], row[i]))

        return tuple(parsed_row)
    
    def _convertData(self, type, data):
        """Convert Crux type to Readable type"""
        if type == 'boolean':
            return bool(data)
        elif type == 'int':
            return int(data)
        elif type == 'long':
            return int(data)
        elif type == 'float' or type == 'double':
            return float(data)
        elif type == 'CruxType:Node' or type == 'GraphNode':
            return Vertex.parseVertexFromJson(data)
        elif type == 'CruxType:Relation' or type == 'GraphRelation':
            return Edge.parseEdgeFromJson(self._graph_schema, data)
        elif type == 'CruxType:Path':
            return Path.parsePathFromJson(self._graph_schema, data)
        elif type.startswith('CruxType:List'):
            return self._parseList(type, data)
        elif type.startswith('CruxType:Map'):
            return self._parseMap(type, data)
        else:
            return data

    def _parseList(self, type, data):
        """Parse 'CruxType:List' type"""
        parsed_data = json.loads(data)
        newType = type[len('CruxType:List') + 1:type.find('>')]

        return [self._convertData(newType, json.dumps(entry)) for entry in parsed_data]

    def _parseMap(self, type, data):
        """Parse 'CruxType:Map' type"""
        parsed_data = json.loads(data)
        newTypes = type[len('CruxType:Map') + 1:-2].split(',')

        result = {}

        for entry in parsed_data.keys():
            key = self._convertData(newTypes[0], entry)
            result[key] = self._convertData(newTypes[1], parsed_data[entry])

        return result

    def toRDD(self, sc, parallelism=1):
        """
        Transform to RDD
        param sc: SparkContext
        param parallelism: RDD parallelism
        """
        return transformToRDD(self, sc, parallelism)

class Connection(object):
    """
    Connection class for connecting to the Stellar database.
    """

    def __init__(
            self, 
            host: str, 
            port: int, 
            auth: Union[str, None] = None, 
            username: Union[str, None] = None, 
            password: Union[str, None] = None,
            kerberos_service_name: Union[str, None] = None,):
        """Connect to HiveServer2

        :param host: What host HiveServer2 runs on
        :param port: What port HiveServer2 runs on. Defaults to 10000.
        :param auth: The value of hive.server2.authentication used by HiveServer2. Defaults to ``NONE``.
        :param username: Use with auth='LDAP' only
        :param password: Use with auth='LDAP' only
        :param kerberos_service_name: Use with auth='KERBEROS' only

        The way to support LDAP and GSSAPI is originated from cloudera/Impyla:
        https://github.com/cloudera/impyla/blob/255b07ed973d47a3395214ed92d35ec0615ebf62
        /impala/_thrift_api.py#L152-L160
        """
        username = username or getpass.getuser()

        if (password is not None) != (auth in ('LDAP', 'CUSTOM')):
            raise ValueError(
                "Password should be set if and only if in LDAP or CUSTOM mode; "
                "Remove password or use one of those modes")
        if (kerberos_service_name is not None) != (auth == 'KERBEROS'):
            raise ValueError(
                "kerberos_service_name should be set if and only if in KERBEROS mode"
            )
        if port is None:
            port = 10000
        if auth is None:
            auth = 'NONE'
        socket = thrift.transport.TSocket.TSocket(host, port)
        if auth == 'NOSASL':
            # NOSASL corresponds to hive.server2.authentication=NOSASL in hive-site.xml
            self._transport = thrift.transport.TTransport.TBufferedTransport(socket)
        elif auth in ('LDAP', 'KERBEROS', 'NONE', 'CUSTOM'):
            import thrift_sasl

            if auth == 'KERBEROS':
                # KERBEROS mode in hive.server2.authentication is GSSAPI in SASL
                sasl_auth = 'GSSAPI'
            else:
                sasl_auth = 'PLAIN'
                if password is None:
                    # Password doesn't matter in NONE mode, just needs to be nonempty.
                    password = 'x'

            def sasl_factory():
                if sasl_auth == 'GSSAPI':
                    sasl_client = PureSASLClient(host, mechanism="GSSAPI", service=kerberos_service_name)
                elif sasl_auth == 'PLAIN':
                    sasl_client = PureSASLClient(host, mechanism="PLAIN", username=username, password=password)
                else:
                    raise AssertionError("Unsupported SASL mechanism")
                return sasl_client

            self._transport = thrift_sasl.TSaslClientTransport(sasl_factory, sasl_auth, socket)
        else:
            # All HS2 config options:
            # https://cwiki.apache.org/confluence/display/Hive/Setting+Up+HiveServer2#SettingUpHiveServer2-Configuration
            # PAM currently left to end user via thrift_transport option.
            raise NotImplementedError(
                "Only NONE, NOSASL, LDAP, KERBEROS, CUSTOM "
                "authentication are supported, got {}".format(auth))
        protocol = thrift.protocol.TBinaryProtocol.TBinaryProtocol(self._transport)
        self._client = TCLIService.Client(protocol)
        # oldest version that still contains features we care about
        # "V6 uses binary type for binary payload (was string) and uses columnar result set"
        protocol_version = ttypes.TProtocolVersion.HIVE_CLI_SERVICE_PROTOCOL_V6
        try:
            self._transport.open()
            open_session_req = ttypes.TOpenSessionReq(
                client_protocol=protocol_version,
                configuration={},
                username=username,
            )
            response = self._client.OpenSession(open_session_req)
            hive._check_status(response)
            assert response.sessionHandle is not None, "Expected a session from OpenSession"
            self._sessionHandle = response.sessionHandle
            assert response.serverProtocolVersion == protocol_version, \
                "Unable to handle protocol version {}".format(response.serverProtocolVersion)
            
        except:
            self._transport.close()
            raise

    def __enter__(self):
        """Transport should already be opened by __init__"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Call close"""
        self.close()

    @property
    def client(self):
        return self._client

    @property
    def sessionHandle(self):
        return self._sessionHandle

    def close(self):
        req = ttypes.TCloseSessionReq(sessionHandle=self._sessionHandle)
        response = self._client.CloseSession(req)
        self._transport.close()
        hive._check_status(response)

    def execute(self, operation: str, graph_schema: Union[GraphSchema, None] = None) -> CypherResult:
        """
        Execute a query on the database.

        :param operation: The query to execute
        """
        cursor = CypherResult(self, graph_schema)
        cursor.execute(operation)
        return cursor
        


#!/usr/bin/env python

import platform
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import versioneer


# class PyTest(TestCommand):
#     def finalize_options(self):
#         TestCommand.finalize_options(self)
#         self.test_args = []
#         self.test_suite = True

#     def run_tests(self):
#         # import here, cause outside the eggs aren't loaded
#         import pytest
#         errno = pytest.main(self.test_args)
#         sys.exit(errno)


with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name="PyStellarDB",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Python interface to StellarDB",
    long_description=long_description,
    url='https://github.com/WarpCloud/PyStellarDB',
    author="Zhiping Wang",
    author_email="zhiping.wang@transwarp.io",
    license="Apache License, Version 2.0",
    python_requires='>=3.6',
    packages=find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Database :: Front-Ends",
    ],
    install_requires=[
        'future',
        'python-dateutil',
        'pyhive[hive_pure_sasl]',
        'thrift>=0.10.0',
    ],
    extras_require={
        'pyspark': ['pyspark>=2.4.0'],
        'kerberos': ['kerberos>=1.3.0'],
    },
    tests_require=[
        'mock>=1.0.0',
        'pytest',
        'pytest-cov',
        'requests>=1.0.0',
        'requests_kerberos>=0.12.0',
    ],
    package_data={
        '': ['*.rst'],
    },
    entry_points={}
)
